"""
Description:

    An experimental plone tool (called portal_speller) for spell checking
    text bodies.

    The tool makes use of external programs to actually do the spell checking.
    There are currently two available speller definitions:

        aspell: Uses the GNU aspell program to do spell checking
        ispell: Uses the GNU ispell program to do spell checking

    Other spellers classes can be added; simply add a definition for the
    speller and call it. See the aspell or ispell definition for what can
    be defined to control the action of the speller.

    A portal tool is created for the ZMI where:

      - speller command options can be edited
      - action can be amended
      - an supplementary wordlist can be edited

"""
# from zLOG import LOG
from Globals import InitializeClass
from AccessControl import Permissions, getSecurityManager, ClassSecurityInfo, ModuleSecurityInfo

from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from OFS.ObjectManager import UNIQUE
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.ActionProviderBase import ActionProviderBase

from PipeCommand import PipeCommand

from interfaces import ISpellCheckTool

from string import split, strip, join

#
# Add new spellers here
#
# try to establish valid spellers
spellers = {}
try:
    import aspell
    spellers['aspell'] = aspell.defs
except:
    pass
try:
    import ispell
    spellers['ispell'] = ispell.defs
except:
    pass

# Default maximum size for texts to check
# Measured in characters
maxsize = 70000

def orderDict(dict):
    """
    A utility function to order a dict into a key,value list
    Used to render language selector alphabetically
    """
    l = []
    for i in dict.keys():
        l += i,
    l.sort()
    ol = ()
    for i in l:
        ol += (i,dict[i]),
    return ol

class CMFSpellCheckTool(SimpleItem, PropertyManager,ActionProviderBase):
    """
    This defines a basic SpellingTool which is implemented as
    an action for any content-type which can be spell checked.
    """
    id = 'portal_speller'
    meta_type = 'Spell Check Tool'
    toolicon = 'spell_icon.gif'

    __implements__ = ( ISpellCheckTool, )

    security = ClassSecurityInfo()
    security.declareObjectProtected(CMFCorePermissions.View)
    __roles__ = ()

    # how and where the SpellChecker will be deployed
    _actions = [ActionInformation(
        id='spelling'
        , title='Spelling'
        , action=Expression(text='string: ${object_url}/spellchecker')
        , condition=Expression(text="python:member and hasattr(object,'EditableBody') and hasattr(object,'edit')")
        , permissions=(CMFCorePermissions.ModifyPortalContent,)
        , category='object_buttons'
        , visible=1
        )]

    manage_options = (
        (
        {'label':'Speller', 'action':'manage_speller',},
        {'label':'Wordlist', 'action':'manage_wordlist',},
        ) + ActionProviderBase.manage_options
        + SimpleItem.manage_options )

    # ------------------------------------------------------------------
    #
    # ZMI Management methods
    #
    # ------------------------------------------------------------------

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_speller')
    manage_speller = PageTemplateFile('www/manageSpeller', globals())

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manage_wordlist')
    manage_wordlist = PageTemplateFile('www/manageWordlist',globals())

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manageSpeller')
    def manageSpeller(self, REQUEST=None):
        """
        Update speller controls
        The ZMI manage function that maintains the control
        settings (command line switches and options) for your selected speller
        --
        Implemented in two parts:
         1 - specify speller to be used; changing this resets ALL switches
         2 - amend command line switches and options
        """
        if REQUEST:
            # An interactive call
            # LOG(__name__,0,'manageSpeller','Presumed interactive call')
            request = REQUEST
            action = request.get('action',None)
            newspeller = request.get('speller',spellers.keys()[0])
            if newspeller!=self.speller and action=='Change speller':
                # change of speller detected
                #LOG(__name__,0,'manageSpeller','Change of speller detected')
                self.speller = newspeller
                # reset options to the default
                self.binpath = spellers[newspeller].get('binpath','/usr/bin/')
                self.dict = spellers[newspeller].get('dict','en')
                self.dictsw = spellers[newspeller].get('dictsw',None)
                self.language = spellers[newspeller].get('language','en')
                self.languagesw = spellers[newspeller].get('languagesw',None)
                self.encoding = spellers[newspeller].get('encoding','utf-8')
                self.encodingsw = spellers[newspeller].get('encodingsw',None)
                self.pipesw = spellers[newspeller].get('pipesw','')
                self.switches = spellers[newspeller].get('switches','')
                self.canreload = spellers[self.speller].get('canreload',None)
                self.wordlistsw = spellers[newspeller].get('wordlistsw',None)
                self.maxsize = maxsize
            if action=='Save Options':
                # the options have changed
                #LOG(__name__,0,'manageSpeller','Interactive Options update')
                self.binpath = request.get('binpath',self.binpath)
                self.dict = request.get('dict',self.dict)
                self.language = request.get('language',self.language)
                self.encoding = request.get('encoding',self.encoding)
                self.pipesw = request.get('pipesw',self.pipesw)
                self.switches = request.get('switches',self.switches)
                self.maxsize = request.get('maxsize',self.maxsize)
                self.wordlistsw = request.get('wordlistsw',self.wordlistsw)
            return REQUEST.RESPONSE.redirect(self.absolute_url() + "/manage_workspace")
        else:
            # Called directly from init (probably)
            # LOG(__name__,0,'manageSpeller','Presumed init call')
            return

    security.declareProtected(CMFCorePermissions.ManagePortal, 'manageWordlist')
    def manageWordlist(self, REQUEST=None):
        """
        Manage the supplementary Wordlist
        The Wordlist is saved as a personal wordlist for the
        user Zope is running as; the location of this file is
        determined by the speller. For some spellers, this file
        can be read by Zope - this feature is controlled with the
        "canreload" switch; if available, it will appear on
        the Wordlist page as an option.

        SystemManager can also upload or download the file outside of Zope or
        edit it directly (in ZMI)

        XXX contains data specific to aspell/ispell for adding words

        Page is rendered as a multipart-form: the "action" indicates which
        submit button was hit
        - Save - changes tp wordlist file specification
        - Save Wordlist - saves the list of words to the file
        - Reload - if available, reloads words from wordlist file
        - Import (not active)
        - Export (not active)
        """
        request = REQUEST
        action = request['action']
        if action=='Save':
            # Save Wordlist switches
            self.wordlistsw = request.get('wordlistsw',self.wordlistsw)
            self.wordlistfile = request.get('wordlistfile','')
        if action=='Save Wordlist':
            # The wordlist was edited in the ZMI so save it please
            # Open a pipe to the speller
            command = self.getSpellWordlistCommand()
            if command:
                pipe = PipeCommand('%s %s' % (self.getSpellBaseCommand(),
                                              command),'traperrors')
                wordlist = split(request['wordlist'],'\n')

                if type(wordlist)==type([]):
                    # loop through the words
                    self.wordlist = []
                    for word in wordlist:
                        w = strip(word)
                        pipe.pumpData('*%s\n' % w)
                        self.wordlist += w,
                else:
                    # just a single word or maybe nothing
                    pipe.pumpData('*%s\n' % wordlist)
                # Tell speller to save the personal wordlist
                pipe.pumpData('#\n')
                pipe.pumpDone()
                # We catch any errors as there may be problems
                # with the file name which can be corrected here
                # They get saved so they can be rendered on the page
                self.wlerrors = pipe.errors()
                # LOG(__name__,0,'Errors',self.wlerrors)
                pipe.close()
        if action=='Reload':
            # The wordlist should be loaded from the filesystem
            # This only works for aspell
            # Open a pipe to the speller
            command = self.getSpellWordlistCommand()
            if command:
                # LOG(__name__,0,'Command',command)
                pipe = PipeCommand('%s %s' % (self.getSpellBaseCommand(),
                                              command),'traperrors')
                pipe.pumpData('$$pp\n')
                pipe.pumpDone()
                readlist = pipe.suckData()
                if readlist:
                    (words,wordlist) = split(readlist[1],":")
                    pipe.suckDone()
                    wordlist = split(wordlist,',')
                    self.wordlist = []
                    for word in wordlist:
                        self.wordlist += strip(word),
                else:
                    self.wordlist = []
                pipe.close()
        if action=='Import':
            # Import data from an uploaded file
            self.wordlist = request['wordlist']
        if action=='Export':
            # Download the wordlist to a file
            self.wordlist = request['wordlist']
        return REQUEST.RESPONSE.redirect(self.absolute_url() + \
                                         "/manage_wordlist")

    def __init__(self):
        """
        Sets a speller to the first available speller if none
        has been defined
        """
        if not hasattr(self,'speller'):
            # Initialise a new instance
            # LOG(__name__,0,'manageSpeller','Initialise instance')
            self.speller = spellers.keys()[0]
            self.title = 'Portal controls for Spell Checking'
            self.binpath = spellers[self.speller].get('binpath','/usr/bin/')
            self.dict = spellers[self.speller].get('dict','en')
            self.dictsw = spellers[self.speller].get('dictsw',None)
            self.language = spellers[self.speller].get('language','en')
            self.languagesw = spellers[self.speller].get('languagesw',None)
            self.encoding = spellers[self.speller].get('encoding','utf-8')
            self.encodingsw = spellers[self.speller].get('encodingsw',None)
            self.pipesw = spellers[self.speller].get('pipesw','')
            self.switches = spellers[self.speller].get('switches','')
            self.canreload = spellers[self.speller].get('canreload',None)
            self.wordlistsw = spellers[self.speller].get('wordlistsw',None)
            self.wordlistfile = ''
            self.canreload = spellers[self.speller].get('canreload',None)
            self.wordlist = []
            self.wlerrors = []
            self.maxsize = maxsize

    security.declareProtected(CMFCorePermissions.ManagePortal, 'validSpellers')
    def validSpellers(self):
        """
        Provide a list of available spellers
        This list is generated from speller definition files that have
        been successfully loaded by this module.
        (into dictionary 'spellers' - see above)
        """
        return (spellers.keys())

    security.declareProtected(CMFCorePermissions.ManagePortal, 'spellerDefaults')
    def spellerDefaults(self):
        """
        Provide a list of parameters for each available speller.
        These are picked up from speller.defs, a dictionary defined
        in a module for each speller
        """
        if self.speller:
            return spellers[self.speller]
        else:
            return spellers[spellers.keys()[0]]

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getCurrentSpeller')
    def getCurrentSpeller(self):
        """
        Returns the name of the current speller; this is the same as the
        name of the executable program
        """
        return self.speller

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellPath')
    def getSpellPath(self):
        """
        Returns the path to the speller executable
        """
        return self.binpath

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellPath')
    def getWLErrors(self):
        """
        Returns a list of errors from Wordlist management
        """
        return getattr(self,'wlerrors',[])

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellDictSw')
    def getSpellDictSw(self):
        """
        Returns the current switch setting for using a specific dictionary
        If this is None, no dictionary will be specified so the spellers
        default (or configured) dictionary will be used
        """
        return self.dictsw

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellDict')
    def getSpellDict(self):
        """
        Returns the currently defined dictionary name
        """
        return self.dict

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellDicts')
    def getSpellDicts(self):
        """
        Returns an ordered list of available dictionary names
        --
        We do not know which dictionaries have been installed;
        SysMan should have edited the list available in the speller defs module
        """
        return orderDict(spellers[self.speller]['dicts'])

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellLanguageSw')
    def getSpellLanguageSw(self):
        """
        Returns the control switch used to specify the language
        used by the speller; the dictionary should already have been installed.
        If no switch is defined, the speller will just use its internal
        notion of a which language to use - some spellers allow this to be
        controlled through a .conf file
        """
        return self.languagesw

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellLanguage')
    def getSpellLanguage(self):
        """
        Returns the currently defined language
        """
        return self.language

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellLanguages')
    def getSpellLanguages(self):
        """
        Returns an ordered list of available languages
        --
        We do not know which languages have been installed;
        SysMan should have edited the list available in the speller defs module
        """
        return orderDict(spellers[self.speller]['languages'])

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellEncodingSw')
    def getSpellEncodingSw(self):
        """
        Returns the current switch for specifying an encoding
        If this is None, no encoding will be specified
        """
        return self.encodingsw

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellEncoding')
    def getSpellEncoding(self):
        """
        Returns the current encoding
        """
        return self.encoding

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellEncodings')
    def getSpellEncodings(self):
        """
        Returns an unordered list of available encodings
        --
        We do not know which encodings are available;
        SysMan should have edited the list available in the speller defs module
        """
        return spellers[self.speller]['encodings']

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellPipeSw')
    def getSpellPipeSw(self):
        """
        Returns current settings for pipe switch
        """
        return self.pipesw

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellSwitches')
    def getSpellSwitches(self):
        """
        Returns current settings for additional switches
        """
        return self.switches

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellWordlistSw')
    def getSpellWordlistSw(self):
        """
        Returns the switch for loading an extra wordlist
        """
        return self.wordlistsw

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellWordlistFile')
    def getSpellWordlistFile(self):
        """
        Returns the path/filename to be used for the personal wordlist
        """
        return self.wordlistfile

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getSpellWordlist')
    def getSpellWordlist(self):
        """
        Returns a wordlist; a list of words to be used as
        a supplement to the spellers standard dictionary
        The speller will get this data delivered as a file (from the
        FileSystem) so this method returns the actual words
        for editing
        """
        return join(self.wordlist,'\n')

    security.declareProtected(CMFCorePermissions.ManagePortal, 'getMaxSize')
    def getMaxSize(self):
        """
        Returns the Maximum Size of text we are prepared to check
        Very large texts cause spellers to hang and in sometimes
        leaves a spinning Zope
        """
        return int(self.maxsize)

    # ------------------------------------------------------------------
    #
    # Class Attributes and Methods
    #
    # ------------------------------------------------------------------

    security.declareProtected('Modify portal content', 'getSpellCommand')
    def getSpellBaseCommand(self):
        """
        Builds the path, executable and pipe switch components to make
        a minimal command
        """
        # Build path to binary; note a / is added to the pathname
        # this may break non-Unix systems
        path = '%s/%s' % (self.getSpellPath(),self.getCurrentSpeller())
        # make a pipe switch if we know how to pass it
        switch = self.getSpellPipeSw()
        if switch:
            pipesw = '%s' % switch
        else:
            pipesw = ''
        spelt = '%s %s' % (path, pipesw)
        return spelt

    security.declareProtected('Modify portal content', 'getSpellCommand')
    def getSpellWordlistCommand(self):
        """
        Build the command component to specify a personal wordlist
        If either wordlist switch or the filename are empty we return nothing
        This prevents confusing the speller especially in the wordlist
        management screen
        """
        # add on wordlist if its not empty
        # and we know how to pass it
        switch = self.getSpellWordlistSw()
        filename = self.getSpellWordlistFile()
        if switch and filename:
            wordlist = '%s %s' % (switch, filename)
        else:
            wordlist = ''
        if wordlist:
            # Make sure we can do this
            spelt = '%s' % wordlist
        else:
            spelt = ''
        return spelt

    security.declareProtected('Modify portal content', 'getSpellCommand')
    def getSpellCommand(self):
        """
        Build the speller command to be called from all the data
        we have been told about; if switches are not specified,
        functions will not be activated.
        --
        starts with the BaseCommand, adds on extra switches and conditionally
        adds personal wordlist: wordlist is ignored if
        - there is no switch
        - there is no filename
        - the wordlist is empty
        """
        # get a basic command
        path = self.getSpellBaseCommand()
        if self.wordlist:
            wordlist = self.getSpellWordlistCommand()
        else:
            wordlist = ''
        # make a dictionary switch if we know how to pass it
        switch = self.getSpellDictSw()
        if switch:
            dict = '%s%s' % (switch,self.getSpellDict())
        else:
            dict = ''
        # make a language switch if we know how to pass it
        switch = self.getSpellLanguageSw()
        if switch:
            language = '%s%s' % (switch,self.getSpellLanguage())
        else:
            language = ''
        # make an encoding switch if we know how to pass it
        switch = self.getSpellEncodingSw()
        if switch:
            encoding = '%s%s' % (switch,self.getSpellEncoding())
        else:
            encoding = ''
        # get the general switches as well
        switches = '%s' % self.getSpellSwitches()
        # and add everything together
        spelt = '%s %s %s %s %s %s' % (path
                                    , wordlist
                                    , dict
                                    , encoding
                                    , language
                                    , switches
                                    )
        # whew!
        # LOG(__name__,0,'Command',spelt)
        return spelt

# Class instanciation
InitializeClass(CMFSpellCheckTool)
