"""
Description:

    A class for creating and utilising a pipe

    A pipe allows zope to maintain a sort of dialogue with a
    command-line tool on your chosen platform. Its not quite
    a fully interactive dialogue because input and output
    occur asynchronously, all input must be finished before
    any output is available

    Typcially a pipe is used like this

    - create a pipe for a command

    - pump data down the pipe as input to the command

    - suck data out of the pipe as output from the command

    - close the pipe

    Note that all input must be finished before output can be
    accessed (this is what a pipe is); the methods *must* be
    called in order. For example

      pipe = PipeCommand(command)
      pipe.pumpData(inputline)
      pipe.pumpData(inputline)
      ...
      pipe.pumpDone()
      output = pipe.suckData()
      pipe.suckDone() # optional if not error trapping
      pipe.close()

    is the same as this command line

      cat inputlines | command > output

    All three elements: command, input and output can be abused
    in whatever way you want even destructively so take care.
    Generally security restrictions imposed by Zope are ignored
    when working with pipes, only your O/S limits what can and
    cannot happen and this is dependent on the user your Zope
    instance is running under.

    Do not implement pipes where Zope runs as root!
    Pipes are a security hazard!

    _YOU HAVE BEEN WARNED_

"""
#import zLOG
from popen2 import Popen3, popen2

try:
    from interfaces import IPipeCommand
except:
    class IPipeCommand: pass

class UPipe:
    """
    Creates a pipe for a command providing methods
    to pump and suck data through it
    Unix version (Popen3 is reported as unavailable under Windows
    """

    def __init__(self,command=None,traperrors=0):
        """
        Open a pipe for a specified command
        --
        command = the command to be executed (with switches/parameters)
        traperrors = true/false - whether errors should be caught for later
          reporting; default is not to catch error messages

        Note: it is better *not* to traperrors unless you have good reason to
              This is because of the nature of pipes and interprocess
              communication which can cause deadlocks - see Python module docs
        """
        self.command = command
        self.traperrors = traperrors
        try:
            self.pipe = Popen3(self.command,self.traperrors)
        except:
            raise

    def __del__(self):
        """
        Properly closes the pipe in an elegant fashion
        """
        if self.pipe.tochild:
            self.pipe.tochild.close()
        if self.pipe.fromchild:
            self.pipe.fromchild.close()
        if self.pipe.childerr:
            self.pipe.childerr.close()
        return self.pipe.wait()

    def getCommand(self):
        """
        Returns the command used to open the pipe
        """
        return self.command

    def pumpData(self,data):
        """
        Pump data down the pipe
        """
        if data:
            self.pipe.tochild.write(data)
        return

    def pumpDone(self):
        """
        All input has been completed
        """
        self.pipe.tochild.close()
        return

    def suckData(self):
        """
        Suck data out of the pipe
        """
        return self.pipe.fromchild.readlines()

    def suckDone(self):
        """
        All output has been recovered
        """
        self.pipe.fromchild.close()
        return

    def status(self):
        """
        Returns the current status of the command
        See popen2 module docs for info
        """
        return self.pipe.poll()

    def errors(self):
        """
        If errors have been trapped, then return them all
        """
        if self.pipe.childerr:
            return self.pipe.childerr.readlines()
        else:
            return

    def close(self):
        """
        Controlled exit
        Same as __del__
        """
        return self.__del__()

class WPipe:
    """
    untested Windows version of the same thing
    --
    These mechanisms tend to leave zombie processes hanging about
    on Unix, probably because the pipe is not getting closed properly
    --
    accesses pipe machinery at a lower level than Popen3
    """

    def __init__(self,command=None):
        """
        Opens a pipe
        """
        self.command = command
        try:
            self.cout,self.cin = popen2(command)
        except:
            raise

    def __del__(self):
        """
        Properly closes the pipe in an elegant fashion
        """
        if self.cin:
            self.cin.close()
        if self.cout:
            self.cout.close()
        # need something else here to splat the Zombie
        self.wait() # might do it
        return

    def getCommand(self):
        """
        Returns the command used to open the pipe
        """
        return self.command

    def pumpData(self,data):
        """
        Pump data down the pipe
        """
        if data:
            print >> self.cin, data
        return

    def pumpDone(self):
        """
        All input has been completed
        """
        self.cin.close()

    def suckData(self):
        """
        Suck data out of the pipe
        """
        return self.cout.readlines()

    def suckDone(self):
        """
        All output has been recovered
        """
        self.cout.close()

    def status(self):
        """
        We have no monitoring function
        """
        return None

    def errors(self):
        """
        We have no error trapping function
        """
        return None

    def close(self):
        """
        Same as __del__
        """
        return self.__del__()

class PipeCommand(UPipe,WPipe):
    """
    Wrapper for Unix/Windows implementations
    """

    def __init__(self,command=None,traperrors=0,ostype='unix'):
        """
        Makes an instance of the appropriate OS pipe
        --
        command = command to be executed
        traperrors = whether to catch errors or not
        ostype = which flavour of pipe to use
          unix (default) or windows
        """
        if ostype=='windows':
            WPipe.__init__(self,command)
        else:
            UPipe.__init__(self,command,traperrors)
