"""
Description:

    A spell check utility class.

    Can be used to spell check "normal" text or html.  For html text,
    the spellers native HTML interpreter is used; this often allows
    embedded text such as "title" or "alt" to be checked.

"""

from Products.CMFCore.utils import getToolByName

from CMFSpellCheckTool import CMFSpellCheckTool
from PipeCommand import PipeCommand
from SpellingMistake import SpellingMistake, InputBuffer

# from zLOG import LOG

import string, re
from types import StringType
from popen2 import popen2, Popen3

def escape(text):
    """
    Escapes characters with a \
    Used to avoid breaking javascript
    """
    return string.replace(text,"'","\'")

def doubleescape(text):
    """
    Escapes characters with a double \; means js gets an escaped
    single quote in a string
    Used to avoid breaking javascript
    """
    return string.replace(text,"'","\\'")

def result_scanner(stream, out):
    """
    Munges the output stream from the command-line tool breaking out
    misspelled words and where they occur (line # and a character offset)
    This construction is the same for aspell and ispell
    --
    Only mistakes should arrive here (using ! leader in spell_check)
    but all if-traps left in for the future - stop the ! to activate
    these switches
    """

    if out is not None:
        out.write('result_scanner\n')

    mistakes = []
    line_count = 0
    spaceMatch = re.compile("(\s+)", re.IGNORECASE | re.DOTALL)

    for l in stream:
        if out is not None:
            out.write('output line: ' + l + '\n')
        s = l[0]
        if s == '\n':
            line_count += 1
        elif s == '*':
            # Word is in dictionary (i.e. it is not misspelled).
            continue
        elif s == '&':
            # Word is not in dictionary and there are suggestions.
            word, number, offset = l[2:l.find(':')].split(' ')
            try:
                # Convert offset to an int and subtract 2 because "^ " was added
                # to the beginning of the line.
                offset = string.atoi(offset) - 2
            except:
                offset = -1
            replacementWords = spaceMatch.sub('', l[l.find(':') + 1:])
            r = replacementWords.split(',')
            replacements = ()
            for i in r:
                replacements += doubleescape(i),
            mistakes.append(SpellingMistake(word, line_count, offset, replacements))
        elif s == '#':
            # Word is not in dictionary and there are no suggestions.
            # Line format: # <misspelled_word> <offset>
            word, offset = l[2:].split(' ')
            try:
                # Convert offset to an int and subtract 2 because "^ " was added
                # to the beginning of the line.
                offset = string.atoi(offset) - 2
            except:
                offset = -1
            # Force the original word thru as the suggestion to make it
            # easier to correct
            mistakes.append(SpellingMistake(word, line_count, offset, (word,)))
        elif s == '@':
            # First line of aspell output; can be ignored.
            continue
        else:
            # Who knows?
            continue
    # zLOG.LOG(__name__,0,'Result of command',mistakes)

    return mistakes


def spell_check(command, text_stream, out = None):
    """
    Does a spell check on the text_stream and returns a list of SpellingMistake
    records representing any misspelled words found.
    """
    #zLOG.LOG(__name__,0,'Text to be checked','echo "%s"|%s' % (text_stream,command))
    lines = []

    if out is not None:
        out.write('spell_check\n')

    # XXX this isn't good enough - twill barf on StringMethods
    if isinstance(text_stream, StringType):
        text_stream = InputBuffer(text_stream)

    # See the PipeCommand module on parameters for this
    # esp. to change action for Windows and other OSs
    pipe = PipeCommand(command)

    # Go faster stripes - Only returns mistakes
    pipe.pumpData('!\n')
    # Loop thru lines of text passing to speller and saving
    # for rendering mistakes in context
    for l in text_stream:
        # Build a list of contextual lines
        #
        # we have to strip the lines to throw away \r which
        # break the js on the SpellCorrector;
        # also need to double escape single quotes
        w = l.rstrip()
        # lines += doubleescape(w),
        lines += w,
        if out is not None:
            out.write('spell_check: line = ' + w + '\n')
        # ^ protects the line from accidental leading chars controlling speller
        # needs a spacer or else replacement gets misplaced
        pipe.pumpData('^ ' + w + '\n')
    pipe.pumpDone()
    # XXX error trapping on pipe should go here
    results = result_scanner(pipe.suckData(), out)
    pipe.close()
    return (results, lines)

class SpellCheckUtility:
    """
    SpellCheckUtility class
    This provides the main interface to the spelling machinery
    """

    id = 'portal_speller_parser'
    meta_type = 'Spell Check Tool Utility'

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, text = '', object = None, out = None):

        self.object = object
        self.out = out

        # assume we will be able to check object
        self.goodtype = 'True'
        
        self.misspelled_info = []
        self.lines = ()

        # We cannot check all types of text - some text is cooked
        # as in ZWiki for example so we catch them and choke
        if type(text)!=type('str'):
            # text is cooked, we don't know how to check it yet
            # zLOG.LOG(__name__,0,'Text type failed ',type(text))
            self.goodtype = None
        else:
            # only attempt a check if we know we might succeed
            # Get a CMFSpellCheckTool.
            self.spellTool = getToolByName(object,'portal_speller')

            # test if the text is too long
            # LOG(__name__,0,'Text length %s max permitted %s',(len(text),self.spellTool.getMaxSize()))
            if len(text) > self.spellTool.getMaxSize():
                self.goodtype = None
            else:
                command = self.spellTool.getSpellCommand()
                # zLOG.LOG(__name__,0,'Command to be executed',command)

                # Spell check the entire content.
                (mistakes, textlines) = spell_check(command, text)
                self.goodtype = 'True'
                self.misspelled_info = mistakes
                self.lines = textlines
                # zLOG.LOG(__name__,0,'Mistakes',self.lines)

    def getTypeTest(self):
        """
        Returns pass or failure of type test
        We just pretend to check content types we don't know how to deal with
        Test this value to determine if a real spell check was done
        """
        return self.goodtype

    def getMisspelledInfo(self):
        """
        Returns a list of misspelled words, where each occurs and suggested
        corrections (if any)
        """
        return self.misspelled_info

    def getLines(self):
        """
        Returns all the text broken into individual lines as a list
        """
        l = []
        for i in self.lines:
            l += doubleescape(i),
        return l

    def highlightWordInContext(self, word, line, offset):
        """
        Returns a line of text containing the word to be corrected
        See the stylesheet fragment for rendering style
        --
        Returned as a list as [pretext,word,postext]
        """
        text = self.lines[line]
        pretext = doubleescape(text[:offset])
        wordlength = len(word)
        postext = doubleescape(text[offset+wordlength:])
        wic = [pretext,doubleescape(word),postext]
        # zLOG.LOG(__name__,0,'Word in Context','%s:: after ::%s' % (wic,escape(wic)))
        return wic
